// https://www.mql5.com/ru/forum/1111/page2336#comment_9660624
//   BestInterval
// #define TICKS_CORRECT_TIME //  ,  ,     

// #define TICKS_ALPARI

#include "Rates.mqh"

class TICKS : public DATA<MqlTick>
{
private:
  static string GetTickFlag( uint tickflag )
  {
    string flag = " " + (string)tickflag;

  #define TICKFLAG_MACRO(A) flag += ((bool)(tickflag & TICK_FLAG_##A)) ? " TICK_FLAG_" + #A : ""; \
                            tickflag -= tickflag & TICK_FLAG_##A;
    TICKFLAG_MACRO(BID)
    TICKFLAG_MACRO(ASK)
    TICKFLAG_MACRO(LAST)
    TICKFLAG_MACRO(VOLUME)
    TICKFLAG_MACRO(BUY)
    TICKFLAG_MACRO(SELL)
  #undef TICKFLAG_MACRO

    if (tickflag)
      flag += " FLAG_UNKNOWN (" + (string)tickflag + ")";

    return(flag);
  }

  static long StringToTimeMsc( const string &Str )
  {
    static MqlDateTime DateTime = {0};

    DateTime.year = Str[0] * 1000 + Str[1] * 100 + Str[2] * 10 + Str[3] - '0' * 1111;
    DateTime.mon = Str[5] * 10 + Str[6] - '0' * 11;
    DateTime.day = Str[8] * 10 + Str[9] - '0' * 11;

    DateTime.hour = Str[11] * 10 + Str[12] - '0' * 11;
    DateTime.min = Str[14] * 10 + Str[15] - '0' * 11;
    DateTime.sec = Str[17] * 10 + Str[18] - '0' * 11;

    return(::StructToTime(DateTime) * (long)1000 + Str[20] * 100 + Str[21] * 10 + Str[22] - '0' * 111);
  }

  static double StringToDouble( const string &Str, uint &Pos )
  {
    static const double TenPow[] = {1, 10, 100, 1000, 10000, 100000, 1000000};

    const uint Size = ::StringLen(Str);

    ulong Res = 0;
    uint point = 0;

    while (Pos < Size)
    {
      const int Digit = Str[Pos];

      if (((Digit >= '0') && (Digit <= '9')))
        Res = Res * 10 + Digit - '0';
      else if (Digit == '.')
        point = Pos + 2;
      else
      {
        Pos++;

        break;
      }

      Pos++;
    }

    //https://www.mql5.com/ru/forum/232030/page6#comment_13921691
    return(point ? Res / TenPow[Pos - point] : Res);
  }

  virtual bool StringToTick( const string &Str, MqlTick &Tick ) const
  {
    Tick.time_msc = TICKS::StringToTimeMsc(Str);
    Tick.time = (datetime)(Tick.time_msc / 1000);

    uint Pos = 24;

    Tick.bid = TICKS::StringToDouble(Str, Pos);
    Tick.ask = TICKS::StringToDouble(Str, Pos);

    return(true);
  }

  static long StringToTimeMsc( const uchar &Bytes[], const uint &Pos )
  {
    static MqlDateTime PrevDate = {0};
    static datetime PrevTime = 0;

    static MqlDateTime DateTime = {0};

    DateTime.year = Bytes[Pos + 0] * 1000 + Bytes[Pos + 1] * 100 + Bytes[Pos + 2] * 10 + Bytes[Pos + 3] - '0' * 1111;
    DateTime.mon = Bytes[Pos + 5] * 10 + Bytes[Pos + 6] - '0' * 11;
    DateTime.day = Bytes[Pos + 8] * 10 + Bytes[Pos + 9] - '0' * 11;

    if ((PrevDate.day != DateTime.day) || (PrevDate.mon != DateTime.mon) || (PrevDate.year != DateTime.year))
    {
      PrevTime = ::StructToTime(DateTime);

      PrevDate = DateTime;
    }

  #ifndef TICKS_ALPARI
    return((PrevTime +
            (Bytes[Pos + 11] * 10 + Bytes[Pos + 12] - '0' * 11) * 3600 +
            (Bytes[Pos + 14] * 10 + Bytes[Pos + 15] - '0' * 11) * 60 +
             Bytes[Pos + 17] * 10 + Bytes[Pos + 18] - '0' * 11) * (long)1000 + Bytes[Pos + 20] * 100 + Bytes[Pos + 21] * 10 + Bytes[Pos + 22] - '0' * 111);
/*    return((PrevTime +
            (Bytes[Pos + 11] * 10 + Bytes[Pos + 12]) * 3600 +
            (Bytes[Pos + 14] * 10 + Bytes[Pos + 15]) * 60 +
             Bytes[Pos + 17] * 10 + Bytes[Pos + 18]) * (long)1000 +
             Bytes[Pos + 20] * 100 + Bytes[Pos + 21] * 10 + Bytes[Pos + 22] - ('0' * 111 + '0' * 11 * 3661 * (long)1000)); */
  #else // TICKS_ALPARI
    return((PrevTime +
            (Bytes[Pos + 11] * 10 + Bytes[Pos + 12] - '0' * 11) * 3600 +
            (Bytes[Pos + 14] * 10 + Bytes[Pos + 15] - '0' * 11) * 60 +
             Bytes[Pos + 17] * 10 + Bytes[Pos + 18] - '0' * 11) * (long)1000); // http://ticks.alpari.org/
  #endif // TICKS_ALPARI
  }

  static double StringToDouble( const uchar &Bytes[], uint &Pos )
  {
    static const double TenPow[] = {1, 10, 100, 1000, 10000, 100000, 1000000};

    const uint Size = ::ArraySize(Bytes);

    ulong Res = 0;
    uint point = 0;

    while (Pos < Size)
    {
      const uchar Digit = Bytes[Pos];

      if (((Digit >= '0') && (Digit <= '9')))
        Res = Res * 10 + Digit - '0';
      else if (Digit == '.')
        point = Pos + 2;
      else
      {
        Pos++;

        break;
      }

      Pos++;
    }

    //https://www.mql5.com/ru/forum/232030/page6#comment_13921691
    return(point ? Res / TenPow[Pos - point] : Res);
  }

  virtual bool StringToTick( const uchar &Bytes[], MqlTick &Tick, uint &Pos ) const
  {
    Tick.time_msc = TICKS::StringToTimeMsc(Bytes, Pos);
    Tick.time = (datetime)(Tick.time_msc / 1000);

  #ifndef TICKS_ALPARI
    Pos += 24;
  #else // TICKS_ALPARI
    // http://ticks.alpari.org/
    Pos += 20;

    TICKS::StringToDouble(Bytes, Pos);
  #endif // TICKS_ALPARI

    Tick.bid = TICKS::StringToDouble(Bytes, Pos);
    Tick.ask = TICKS::StringToDouble(Bytes, Pos);

    return(true);
  }

/*
  // https://www.mql5.com/ru/forum/360632/page3#comment_20258484
  static int Digits2(double x) {
     if (x<0) x=-x;
     if (x>=ULONG_MAX) return 0;
     ulong z=ulong(x);
     ulong y= ulong((x-z)*1e16);
     int N=0;
     do {
        z/=10;
        y/=10;
        N++;
     } while(z!=0);
     if(y==0) return 0;
     if(y%10==9) y++;
     int n=0;
     while (y%10==0) {
        y/=10;
        n++;
     }
     return 16-n-N;
  }
*/
  static int GetDigits( double Price )
  {
    int Res = 0;
    int Count = 0;

    while ((bool)(Price = ::NormalizeDouble(Price - (int)Price, 8)) && (Count < 8))
    {
      Price *= 10;

      Res++;
      Count++;
    }

    if (Count >= 8)
      ::Alert("Error: " + __FILE__ + " " + (string)__LINE__ + ", Count = " + (string)Count + ", Res = " + (string)Res + ", Price = " + (string)Price);

    return(Res);
  }

  static int GetDigits( const MqlTick &Tick )
  {
    return(::MathMax(TICKS::GetDigits(Tick.bid), TICKS::GetDigits(Tick.ask)));
  }

  static int PriceToInteger( const double Price, const double point )
  {
    return((int)(Price / point + 0.1));
  }

  static void Normalize( double &Price, const int digits )
  {
    Price = ::NormalizeDouble(Price, digits);

    return;
  }

  static void Normalize( MqlTick &Tick, const int digits )
  {
    TICKS::Normalize(Tick.bid, digits);
    TICKS::Normalize(Tick.ask,  digits);
    TICKS::Normalize(Tick.last, digits);

    return;
  }

#define DAY (24 * 3600)
#define WEEK 7

  static ENUM_DAY_OF_WEEK TimeDayOfWeek( const datetime time )
  {
    return((ENUM_DAY_OF_WEEK)((time / DAY + THURSDAY) % WEEK));
  }

  static datetime FirstDayWeekMonth( const int Year, const int Month, const int Count = 1, const ENUM_DAY_OF_WEEK DayWeek = SUNDAY )
  {
    const datetime time = (datetime)((string)Year + "-" + (string)Month + "-01");

    return(time + ((WEEK + DayWeek - TICKS::TimeDayOfWeek(time)) % WEEK) * DAY + (Count - 1) * WEEK * DAY);
  }

  static datetime LastDayWeekMonth( const int Year, const int Month, const ENUM_DAY_OF_WEEK DayWeek = SUNDAY )
  {
    const datetime time = (datetime)((string)(Year + (Month == 12)) + "-" + (string)((Month + 1) % 12) + "-01") - DAY;

    return(time - ((WEEK + TICKS::TimeDayOfWeek(time) - DayWeek) % WEEK) * DAY);
  }

#undef WEEK
#undef DAY

  //     Ticks[Pos]: Ticks[Pos - 1].time < time.
  static int GetTimePosLeft( const MqlTick &Ticks[], const datetime time )
  {
    int Left = -1;
    int Right = ::ArraySize(Ticks);

    while (Left < Right - 1)
    {
      const int Middle = (Left + Right) >> 1;

      if (Ticks[Middle].time < time)
        Left = Middle;
      else
        Right = Middle;
    }

    return(Right);
  }

  //     Ticks[Pos]: Ticks[Pos].time <= time.
  static int GetTimePosRight( const MqlTick &Ticks[], const datetime time )
  {
    int Left = -1;
    int Right = ::ArraySize(Ticks);

    while (Left < Right - 1)
    {
      const int Middle = (Left + Right) >> 1;

      if (Ticks[Middle].time <= time)
        Left = Middle;
      else
        Right = Middle;
    }

    return(Right - 1);
  }

  static int GetYear( const datetime time )
  {
    MqlDateTime DateTime;

    ::TimeToStruct(time, DateTime);

    return(DateTime.year);
  }

  static void ChangeTime( MqlTick &Tick, const int &AddTime )
  {
    Tick.time_msc += AddTime;
    Tick.time = (datetime)(Tick.time_msc / 1000);

    return;
  }

  static void ChangeTime( MqlTick &Ticks[], const datetime From, const datetime To, int AddTime = 3600 )
  {
    AddTime *= 1000;

    const int PosLeft = TICKS::GetTimePosLeft(Ticks, From);
    const int PosRight = TICKS::GetTimePosRight(Ticks, To);

    for (int i = PosLeft; i <= PosRight; i++)
      TICKS::ChangeTime(Ticks[i], AddTime);

    return;
  }

public:
#define TOSTRING(A) " " + #A + " = " + (string)Tick.A

  static string ToString( const MqlTick &Tick )
  {
    return(TOSTRING(time) + "." + ::IntegerToString(Tick.time_msc % 1000, 3, '0') +
           TOSTRING(bid) + TOSTRING(ask) + TOSTRING(last)+ TOSTRING(volume) + TICKS::GetTickFlag(Tick.flags));
  }
#define TOSTRING2(A) " " + #A + " = " + ::DoubleToString(Tick.A, digits)

  static string ToString( const MqlTick &Tick, const int digits )
  {
    return(TOSTRING(time) + "." + ::IntegerToString(Tick.time_msc % 1000, 3, '0') +
           TOSTRING2(bid) + TOSTRING2(ask) + TOSTRING2(last)+ TOSTRING(volume) + TICKS::GetTickFlag(Tick.flags));
  }

#undef TOSTRING2

#undef TOSTRING

  //        ,
  //           .
  //        NOT_EQUAL_PRICES.
  int FilterTicks( const int MinPips = 0, const int BarInterval = 0 )
  {
    MqlTick MinMax = {0};
    int Amount = 0;

    const double point = ::MathPow(10, -this.GetDigits());

    const int Size = ::ArraySize(this.Data);
    bool FlagUp = true;

    int PriceMinMax = INT_MIN;

    datetime NextBarTime = 0;
    MqlTick TickOpen = {};

    for (int i = 0; i < Size; i++)
    {
      const int PriceBid = TICKS::PriceToInteger(this.Data[i].bid, point);
      const int PriceAsk = TICKS::PriceToInteger(this.Data[i].ask, point);

      bool FlagWrite = false;

      // https://www.mql5.com/ru/forum/318657/page19#comment_13767500
//      if (PriceBid <= PriceAsk) //     ,    .
      {
        if (FlagUp)
        {
          if (PriceBid > PriceMinMax)
          {
            PriceMinMax = PriceBid;

            this.Data[Amount++] = this.Data[i];
            FlagWrite = true;
          }
          else if (PriceAsk <= PriceMinMax - MinPips)
          {
            PriceMinMax = PriceAsk;
            FlagUp = false;

            this.Data[Amount++] = this.Data[i];
            FlagWrite = true;
          }
        }
        else if (PriceAsk < PriceMinMax)
        {
          PriceMinMax = PriceAsk;

          this.Data[Amount++] = this.Data[i];
          FlagWrite = true;
        }
        else if (PriceBid >= PriceMinMax + MinPips)
        {
          PriceMinMax = PriceBid;
          FlagUp = true;

          this.Data[Amount++] = this.Data[i];
          FlagWrite = true;
        }

        if (BarInterval)
        {
          if ((this.Data[i].time >= NextBarTime))
          {
            NextBarTime = (this.Data[i].time / BarInterval + 1) * BarInterval;
            TickOpen = this.Data[i];

            if (!FlagWrite)
              this.Data[Amount++] = this.Data[i];
          }
          else
          {
            const bool ResBid = (this.Data[i].bid > TickOpen.bid);
            const bool ResAsk = (this.Data[i].ask < TickOpen.ask);

            if (ResBid)
              TickOpen.bid = this.Data[i].bid;

            if (ResAsk)
              TickOpen.ask = this.Data[i].ask;

            if ((ResBid || ResAsk) && !FlagWrite)
              this.Data[Amount++] = this.Data[i];
          }
        }
      }
    }

    return(::ArrayResize(this.Data, Amount));
  }

/*
  static int AddLastPrices( MqlTick &Ticks[] )
  {
    const int Last = ::ArraySize(Ticks) - 1;
    int Res =  (Last >= 0) ? 1 : 0;

    Ticks[Last].last = (Ticks[Last].bid + Ticks[Last].ask) / 2;

    for (int i = 0; i < Last; i++)
      if (Ticks[i].time / (24 * 3600) != Ticks[i + 1].time / (24 * 3600))
      {
        Ticks[i].last = (Ticks[i].bid + Ticks[i].ask) / 2;

        Res++;
      }

    return(Res);
  }
*/
  static int AddLastPrices( MqlTick &Ticks[] )
  {
    const int Size = ::ArraySize(Ticks);

    for (int i = 0; i < Size; i++)
      Ticks[i].last = (Ticks[i].bid + Ticks[i].ask) / 2;

    return(Size);
  }

  int AddLastPrices( void )
  {
    return(TICKS::AddLastPrices(this.Data));
  }

#define DAY (24 * 3600)
#define MINUTE 60
  static int AddFirstTicks( MqlTick &Ticks[], const int MinInterval = 7 * DAY )
  {
    const int Size = ::ArraySize(Ticks);

    if ((Size > 1) && (Ticks[Size - 1].time - Ticks[0].time < MinInterval))
    {
      MqlTick BarTicks[];

      const int AddBars = ::ArrayResize(BarTicks, MinInterval / MINUTE);

      for (int i = 0; i < AddBars; i++)
      {
        BarTicks[i] = Ticks[0];

        BarTicks[i].time -= (AddBars - i) * 60;
        BarTicks[i].time_msc = BarTicks[i].time * 1000;
      }

      ::ArrayInsert(Ticks, BarTicks, 0);
    }

    return(::ArraySize(Ticks) - Size);
  }

  int AddFirstTicks( const int MinInterval = 7 * DAY )
  {
    return(TICKS::AddFirstTicks(this.Data, MinInterval));
  }
#undef MINUTE
#undef DAY

  void AntiDST( void )
  {
    const uint Size = this.GetAmount();

    if (Size > 0)
    {
      const int YearFrom = TICKS::GetYear(this.Data[0].time);
      const int YearTo = TICKS::GetYear(this.Data[Size - 1].time);

      for (int Year = YearFrom; Year <= YearTo; Year++)
      {
        TICKS::ChangeTime(this.Data, TICKS::FirstDayWeekMonth(Year, 3, 2), TICKS::LastDayWeekMonth(Year, 3));

        TICKS::ChangeTime(this.Data, TICKS::LastDayWeekMonth(Year, 10), TICKS::FirstDayWeekMonth(Year, 11));
      }
    }

    return;
  }

  static int Correct( MqlTick &Ticks[] )
  {
    int Res = 0;
    const int Size = ::ArraySize(Ticks);

    if (Size)
    {
      Ticks[0].flags = TICK_FLAG_BID | TICK_FLAG_ASK;

      for (int i = 1; _CS(i < Size); i++)
      {
        Ticks[i].flags = ((Ticks[i].bid == Ticks[i - 1].bid) ? 0 : TICK_FLAG_BID) |
                         ((Ticks[i].ask == Ticks[i - 1].ask) ? 0 : TICK_FLAG_ASK) |
                         ((Ticks[i].last == Ticks[i - 1].last) ? 0 : TICK_FLAG_LAST);

        if (Ticks[i].time_msc < Ticks[i - 1].time_msc)
        {
//          _P(TICKS::ToString(Ticks[i - 1])); // #include <Debug.mqh>
//          _P(TICKS::ToString(Ticks[i]));

          Ticks[i].time_msc = Ticks[i - 1].time_msc;
          Ticks[i].time = Ticks[i - 1].time;

          Res++;
        }

      }
    }

  #ifdef TICKS_CORRECT_TIME
    for (int i = 0; _CS(i < Size); i++)
      if (!(Ticks[i].time_msc % 1000))
      {
        Ticks[i].time_msc++;

        Res++;
      }
  #endif // TICKS_CORRECT_TIME

    return(Res);
  }

  uint Add( const string &Str )
  {
    MqlTick Tick = {0};

    if (this.StringToTick(Str, Tick))
      DATA<MqlTick>::Add(Tick); // https://www.mql5.com/ru/forum/1111/page2195#comment_7119365

    return(this.GetAmount());
  }

  uint Add( const uchar &Bytes[], const uint StartPos = 0 )
  {
    const uint Size = ::ArraySize(Bytes);
    bool FirstTick = true;

    MqlTick Tick = {0};

  #ifndef TICKS_ALPARI
    for (uint i = StartPos; i < Size; i++)
  #else // TICKS_ALPARI
    for (uint i = StartPos; i < Size;) // http://ticks.alpari.org/
  #endif // TICKS_ALPARI
    {
      TICKS::StringToTick(Bytes, Tick, i);

      if (FirstTick)
      {
      #ifndef TICKS_ALPARI
        this.SetReserve((int)((Size - StartPos) / (i - StartPos + 1)));
      #else // TICKS_ALPARI
        this.SetReserve((int)((Size - StartPos) / (i - StartPos))); // http://ticks.alpari.org/
      #endif // TICKS_ALPARI

        FirstTick = false;
      }

      DATA<MqlTick>::Add(Tick);
    }

    return(this.GetAmount());
  }

  uint Add( const string &Str[] )
  {
    const int Size = this.SetReserve(::ArraySize(Str));

    for (int i = 0; _CS(i < Size); i++)
      this.Add(Str[i]);

    return(this.GetAmount());
  }

  uint Add( const MqlTick &Ticks[] )
  {
    ::ArrayCopy(this.Data, Ticks, ::ArraySize(this.Data));

    return(this.GetAmount());
  }

  int GetDigits( const uint Amount = 1000 ) const
  {
    int Res = 0;
    MqlTick Tick = {0};

    const uint Size = this.GetAmount();

    if ((!Amount) || (Amount >= Size))
      for (uint i = 0; _CS(i < Size); i++)
      {
        const MqlTick TmpTick = this.Data[i];
        const int digits = TICKS::GetDigits(TmpTick);

        if (digits > Res)
        {
          Res = digits;

          Tick = TmpTick;
        }
      }
    else
      for (uint i = 0; _CS(i < Amount); i++)
      {
        const MqlTick TmpTick = this.Data[(int)(::MathRand() * (long)Size / (SHORT_MAX + 1))];
        const int digits = TICKS::GetDigits(TmpTick);

        if (digits > Res)
        {
          Res = digits;

          Tick = TmpTick;
        }
      }

    ::Print("Digits = " + (string)Res + ": " + TICKS::ToString(Tick, Res));

    return(Res);
  }

  void Normalize( const int digits )
  {
    const uint Size = this.GetAmount();

    for (uint i = 0; _CS(i < Size); i++)
      TICKS::Normalize(this.Data[i], digits);

    return;
  }

  int ToSymbol( const SYMBOL* const CustomSymbol ) const
  {
    return(CustomSymbol.CloneTicks(this.Data));
  }

  int Correct( void )
  {
    return(TICKS::Correct(this.Data));
  }

  uint ToRates( RATES* const Rates ) const
  {
    return(Rates.Add(this.Data));
  }

#define DAY (24 * 3600)
#define MINUTE 60

  int GetSessions( datetime &From[], datetime &To[] ) const
  {
    const int Size = ::ArrayResize(From, ::ArrayResize(To, 7));
    const uint Total = this.GetAmount();

    for (int i = 0; i < Size; i++)
    {
      From[i] = DAY;
      To[i] = 0;
    }

    if (Total)
    {
      int WeekDay = (int)((this.Data[0].time/ DAY + 4) % 7);
      int PrevDay = (int)(this.Data[0].time / DAY);

      for (uint i = 0; i < Total; i++)
      {
        const datetime time = ((this.Data[i].time % DAY) / MINUTE) * MINUTE;
        const int Day = (int)(this.Data[i].time / DAY);

        if (Day != PrevDay)
        {
          WeekDay = (WeekDay + (Day - PrevDay)) % Size;

          PrevDay = Day;
        }

        if (time < From[WeekDay])
          From[WeekDay] = time;

        if (time + MINUTE > To[WeekDay])
          To[WeekDay] = time + MINUTE;
      }
    }

    return(Size);
  }
#undef MINUTE
#undef DAY
};